// Copyright (C) 2004-2007 Tord Romstad

// This file is part of Glaurung.
//
// Glaurung is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//  
// Glaurung is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Glaurung; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


////
//// Includes
////

#include <cassert>
#include <fstream>
#include <iostream>

#include "evaluate.h"
#include "mersenne.h"
#include "movegen.h"
#include "movepick.h"
#include "search.h"
#include "test.h"


////
//// Local definitions
////

namespace {

  /// Functions
  int compare(const void* a, const void* b);
  Color random_color();
  Square random_square();
  Square random_empty_square(const Position& pos);
  PieceType random_promoted_piece_type();

}

////
//// Functions
////

/// Test::perft is a function for debugging the move generator.  It takes
/// a position and a search depth as input, computes the complete game tree
/// to the given depth (disregarding repetition and 50 move draws), and counts
/// the number of nodes visited.  For instance, from the normal starting
/// position, perft(1) = 20, while perft(2) = 400.  

uint64 Test::perft(Position& pos, int depth) {
  assert(pos.is_ok());
  assert(depth >= 0);

  if(depth == 0) return 1ULL;
  else {
    uint64 result = 0ULL;
    MoveStack moves[256];
    UndoInfo u;
    int i, n;

    if(pos.is_check()) {
      n = generate_evasions(pos, moves);
      if(depth == 1)
        return n;
      else {
        Bitboard dc = pos.discovered_check_candidates(pos.side_to_move());
        for(i = 0; i < n; i++) {
          pos.do_move(moves[i].move, u, dc);
          result += Test::perft(pos, depth-1);
          pos.undo_move(moves[i].move, u);
        }
      }
    }
    else {
      Bitboard pinned = pos.pinned_pieces(pos.side_to_move());
      Bitboard dc = EmptyBoardBB;
      if(depth > 1) dc = pos.discovered_check_candidates(pos.side_to_move());
      n = generate_captures(pos, moves);
      for(i = 0; i < n; i++)
        if(pos.move_is_legal(moves[i].move, pinned)) {
          if(depth == 1) result++;
          else {
            pos.do_move(moves[i].move, u, dc);
            result += Test::perft(pos, depth-1);
            pos.undo_move(moves[i].move, u);
          }
        }
      n = generate_noncaptures(pos, moves);
      for(i = 0; i < n; i++)
        if(pos.move_is_legal(moves[i].move, pinned)) {
          if(depth == 1) result++;
          else {
            pos.do_move(moves[i].move, u, dc);
            result += Test::perft(pos, depth-1);
            pos.undo_move(moves[i].move, u);
          }
        }
    }
    return result;
  }
}
            

/// Test::divide_perft is similar to Test::perft, but also prints the number
/// of nodes in the subtree below each root move.  This makes it easy to
/// locate the problem when Test::perft yields an incorrect result.  By
/// comparing the divide_perft output for a buggy and a non-buggy program
/// version, we can easily "zoom in" on the position where an incorrect number
/// of legal moves is generated.

void Test::divide_perft(Position& pos, int depth) {
  MoveStack moves[256], legalMoves[256];
  int numOfMoves = 0, i;
  UndoInfo u;
  uint64 result, acc = 0ULL;

  if(pos.is_check()) 
    numOfMoves = generate_evasions(pos, legalMoves);
  else {
    int n;
    n = generate_captures(pos, moves);
    n += generate_noncaptures(pos, moves + n);
    for(i = 0; i < n; i++)
      if(pos.move_is_legal(moves[i].move))
        legalMoves[numOfMoves++].move = moves[i].move;
  }

  qsort(legalMoves, numOfMoves, sizeof(MoveStack), compare);
  
  for(i = 0; i < numOfMoves; i++) {
    std::cout << move_to_string(legalMoves[i].move) << " ";
    pos.do_move(legalMoves[i].move, u);
    result = Test::perft(pos, depth - 1);
    pos.undo_move(legalMoves[i].move, u);
    acc += result;
    std::cout << result << " " << acc << std::endl;
  }
}


/// Test::random_position sets up a pseudo-random position, for use when
/// debugging.  This function has some serious limitations, and is not
/// quite general enough.  It never returns a position with an en passant
/// square or where there is an en passant capture.  It should also be
/// possible to ask for positions where all or most of the pieces are still
/// on the board, and positions where most of the pices are off the board,
/// in order to get nice random test cases for all phases of the game.
/// FIXME

void Test::random_position(Position& pos) {
  while(true) {
    pos.clear();

    // Place pawns:
    for(Color c = WHITE; c <= BLACK; c++) {
      for(int i = 0; i < 8; i++) {
        Square sq;

        // Is the pawn captures?
        if(genrand_int32() % 2 == 0) continue;

        // Pick square:
        sq = random_empty_square(pos);

        // Is the pawn promoted?
        if(genrand_int32() % 16 == 0)
          pos.put_piece(piece_of_color_and_type(c,
                                                random_promoted_piece_type()),
                        sq);
        else if(square_rank(sq) != RANK_1 && square_rank(sq) != RANK_8)
          pos.put_piece(pawn_of_color(c), sq);
      }
    }

    // Place knights, bishops, rooks and queens:
    for(PieceType pt = KNIGHT; pt <= QUEEN; pt++) {
      for(Color c = WHITE; c <= BLACK; c++) {
        for(int i = 0; i < ((pt == QUEEN)? 1 : 2); i++) {
          Square sq;
          
          // Is the piece captured?
          if(genrand_int32() % 2 == 0) continue;
          
          // Pick square:
          sq = random_empty_square(pos);
          pos.put_piece(piece_of_color_and_type(c, pt), sq);
        }
      }
    }

    // Pick side to move
    pos.sideToMove = random_color();

    // Place kings
    pos.put_piece(king_of_color(WHITE), random_empty_square(pos));
    pos.put_piece(king_of_color(BLACK), random_empty_square(pos));

    // Legal position?
    pos.find_checkers();
    if(!pos.square_is_attacked(pos.king_square(opposite_color(pos.side_to_move())),
                               pos.side_to_move()) &&
       count_1s(pos.checkers()) <= 2)
      // Legal position, exit loop
      break;
  }

  // TODO: EP and castling!

  pos.key = pos.compute_key();
  pos.pawnKey = pos.compute_pawn_key();
  pos.materialKey = pos.compute_material_key();
  pos.mgValue = pos.compute_mg_value();
  pos.egValue = pos.compute_eg_value();
  pos.npMaterial[WHITE] = pos.compute_non_pawn_material(WHITE);
  pos.npMaterial[BLACK] = pos.compute_non_pawn_material(BLACK);
}


/// Test::random_legal_move returns a random legal move for the given
/// position.

Move Test::random_legal_move(const Position& pos) {
  MoveStack mlist[256];
  int n;
  n = generate_legal_moves(pos, mlist);
  if(n == 0)
    return MOVE_NONE;
  else
    return mlist[genrand_int32() % n].move;
}


/// Test::random_pseudo_move returns a "pseudo move", which is something
/// that looks vaguely like a move, but is almost certainly illegal, to the
/// rest of the program.  The returned move consists of two randomly chosen
/// source and destination squares.  Note that it is possible that the returned
/// move will not be legal in *any* position, we could get "moves" like
/// g2e8.  The purpose of this function is to help us test and debug the
/// MovePicker class, which has to handle the case when it receives killer
/// moves which are not pseudo-legal.

Move Test::random_pseudo_move() {
  return make_move(random_square(), random_square());
}


/// Test::random_maybe_legal_move randomly returns a legal move, a random
/// pseudo-move, or MOVE_NONE.  Used in the regression test function for
/// the MovePicker class.

Move Test::random_maybe_legal_move(const Position& pos) {
  switch(genrand_int32() % 8) {
  case 0: case 1: case 2: case 3:
    return random_legal_move(pos);
  case 4: case 5:
    return random_pseudo_move();
  default:
    return MOVE_NONE;
  }
}


/// Test::flip_position makes a "mirror image" of a position.  Used for
/// evaluation symmetry tests.

void Test::flip_position(Position& dst, const Position& src) {
  assert(src.is_ok());
  
  dst.clear();

  // Board
  for(Square s = SQ_A1; s <= SQ_H8; s++)
    if(!src.square_is_empty(s))
      dst.put_piece(Piece(int(src.piece_at(s)) ^ 8), flip_square(s));
    
  // Side to move
  dst.sideToMove = opposite_color(src.sideToMove);

  // Castling rights
  if(src.can_castle_kingside(WHITE)) dst.allow_oo(BLACK);
  if(src.can_castle_queenside(WHITE)) dst.allow_ooo(BLACK);
  if(src.can_castle_kingside(BLACK)) dst.allow_oo(WHITE);
  if(src.can_castle_queenside(BLACK)) dst.allow_ooo(WHITE);

  dst.initialKFile = src.initialKFile;
  dst.initialKRFile = src.initialKRFile;
  dst.initialQRFile = src.initialQRFile;

  for(Square sq = SQ_A1; sq <= SQ_H8; sq++)
    dst.castleRightsMask[sq] = ALL_CASTLES;
  dst.castleRightsMask[make_square(dst.initialKFile, RANK_1)] ^= 
    (WHITE_OO|WHITE_OOO);
  dst.castleRightsMask[make_square(dst.initialKFile, RANK_8)] ^= 
    (BLACK_OO|BLACK_OOO);
  dst.castleRightsMask[make_square(dst.initialKRFile, RANK_1)] ^= WHITE_OO;
  dst.castleRightsMask[make_square(dst.initialKRFile, RANK_8)] ^= BLACK_OO;
  dst.castleRightsMask[make_square(dst.initialQRFile, RANK_1)] ^= WHITE_OOO;
  dst.castleRightsMask[make_square(dst.initialQRFile, RANK_8)] ^= BLACK_OOO;

  // En passant square
  if(src.epSquare != SQ_NONE)
    dst.epSquare = flip_square(src.epSquare);
  
  dst.find_checkers();

  dst.key = dst.compute_key();
  dst.pawnKey = dst.compute_pawn_key();
  dst.materialKey = dst.compute_material_key();
  dst.mgValue = dst.compute_mg_value();
  dst.egValue = dst.compute_eg_value();
  dst.npMaterial[WHITE] = dst.compute_non_pawn_material(WHITE);
  dst.npMaterial[BLACK] = dst.compute_non_pawn_material(BLACK);

  assert(dst.is_ok());
}
  

/// Test::test_move_picker() tests the MovePicker class with feeding it
/// lots of randomly generated positions with or without transposition table
/// moves, and with killer moves which may or may not be pseudo-legal.  For
/// each position, we verify that all moves which should be generated are
/// returned exactly once, and that the move ordering looks correct.  The
/// function returns 'true' when the test succeeds, and 'false' if not.

bool Test::test_move_picker() {
  Position pos;
  MoveStack mlist[256];
  Move ttMove, mateKiller, killer1, killer2;
  int numOfMoves, i, j, n;
  bool pvNode;

  for(i = 0; i < 1000000; i++) {
    // Pick a new randomly generated position
    Test::random_position(pos);

    // Generate legal moves
    numOfMoves = generate_legal_moves(pos, mlist);
    if(numOfMoves == 0) continue;

    // HACK: Use the 'score' slot in the MoveStack objects to keep track of
    // whether the moves have been picked:
    for(j = 0; j < numOfMoves; j++)
      mlist[j].score = 0;

    // Initialize transposition table move and killers:
    if(genrand_int32() % 2)
      ttMove = random_legal_move(pos);
    else
      ttMove = MOVE_NONE;
    mateKiller = Test::random_maybe_legal_move(pos);
    killer1 = Test::random_maybe_legal_move(pos);
    killer2 = Test::random_maybe_legal_move(pos);

    // PV node?
    pvNode = genrand_int32() % 2;

    // Initialise a move picker object:
    MovePicker mp = MovePicker(pos, pvNode, ttMove, mateKiller, killer1,
                               killer2, Depth(1));
    Move m;

    // pos.print();
    
    // Use mp to pick one move at the time:
    for(m = mp.get_next_move(), n = 0; m != MOVE_NONE;
        m = mp.get_next_move(), n++) {
      // std::cout << "Picked " << move_to_string(m) << std::endl;

      // Find the move in mlist[]
      for(j = 0; j < numOfMoves; j++)
        if(m == mlist[j].move) {
          // Has this move been picked before?
          if(mlist[j].score > 0) {
            pos.print();
            std::cout << "In Test::test_move_picker: The move "
                      << move_to_string(m)
                      << " was picked more than once."
                      << std::endl;
            return false;
          }
          else mlist[j].score = n + 1;
          // If we have a TT move, it should be picked first:
          if(ttMove && n == 0 && ttMove != m) {
            pos.print();
            std::cout << "In Test::test_move_picker: TT move was "
                      << move_to_string(ttMove)
                      << ", but first move picked was "
                      << move_to_string(m)
                      << std::endl;
            return false;
          }
          // We could have additional move ordering sanity checks here,
          // but it would be somewhat cumbersome to have to update this
          // function every time we change the move ordering heuristics.

          break;
        }
      if(j == numOfMoves) {
        pos.print();
        std::cout << "In Test::test_move_picker: mp.get_next_move() returned "
                  << move_to_string(m)
                  << ", which is not a legal move."
                  << std::endl;
        return false;
      }
    }
    if(n != numOfMoves) {
      pos.print();
      std::cout << "In Test::test_move_picker: mp.get_next_move() returned "
                << n << " moves, but there are " << numOfMoves
                << " legal moves." << std::endl;
      return false;
    }
  }
  return true;
}


/// Test::test_eval_symmetry tests for symmetry bugs in the evaluation function
/// by evaluating random positions and verifying that the eval of the position
/// and of the corresponding mirrored positioin are equal.

bool Test::test_eval_symmetry() {
  Position p1, p2;
  for(int i = 0; i < 1000000; i++) {
    random_position(p1);
    flip_position(p2, p1);
    if(evaluate(p1, -VALUE_INFINITE, VALUE_INFINITE, 0) !=
       evaluate(p2, -VALUE_INFINITE, VALUE_INFINITE, 0)) {
      std::cout << std::endl;
      p1.print();
      p2.print();
      std::cout << "In Test::test_eval_symmetry: Symmetry test failed."
                << std::endl
                << "evaluate(p1) = "
                << evaluate(p1, -VALUE_INFINITE, VALUE_INFINITE, 0)
                << ", evaluate(p2) = "
                << evaluate(p2, -VALUE_INFINITE, VALUE_INFINITE, 0)
                << std::endl;
      return false;
    }
  }
  return true;
}


/// Test::test_see runs a regression test for the static exchange evaluator.

bool Test::test_see() {
  Position p;
  Move m;
  std::ofstream f;
  
  f.open("see.txt");
  for(int i = 0; i < 100000; i++) {
    random_position(p);
    m = random_legal_move(p);
    if(m == MOVE_NONE) continue;
    f << p.to_fen() << "; " << move_to_string(m) << "; " << p.see(m)
      << std::endl;
  }
  f.close();
  return true;
}
  

/// Test::run_all_tests runs all regression tests.  It is called when
/// Glaurung is started from the command line with the "test" parameter.

bool Test::run_all_tests() {
  bool success = true;

  std::cout << "Testing evaluation symmetry... ";
  if(test_eval_symmetry())
    std::cout << "Success!" << std::endl;
  else {
    std::cout << "***FAILED***" << std::endl;
    std::cout << "Did you remember to set GrainSize to 1 in evaluate.cpp?"
              << std::endl;
    success = false;
  }

  // test_move_picker currently does not work, disabled.
  /*
  std::cout << "Testing MovePicker class... ";
  if(test_move_picker())
    std::cout << "Success!" << std::endl;
  else {
    std::cout << "***FAILED***" << std::endl;
    success = false;
  }
  */

  // test_see();
  
  if(success)
    std::cout << "Alles gute" << std::endl;
  else
    std::cout << "Some tests failed." << std::endl;
  
  return success;
}


void Test::move_ordering_test() {
  FILE* f = fopen("fen-list.txt", "r");

  if(f == NULL) {
    std::cerr << "fen-list.txt not found." << std::endl;
    exit(EXIT_FAILURE);
  }

  char str[128];
  Position p;
  uint64 nodeSum = 0ULL;
  double nodeProduct = 1.0;
  int count = 0;
  
  while(fgets(str, 128, f)) {
    count++;
    p.from_fen(str);
    p.print();
    think(p, true, false, 0, 0, 0, 12, 0);
    nodeSum += nodes();
    nodeProduct *= double(nodes()) / 1000000.0;
    std::cout << count << "positions finished." << std::endl;
    std::cout << "nodeSum = " << nodeSum << ", nodeProduct = " << nodeProduct
              << std::endl;
  }
  std::cout << "sum: " << nodeSum << " product: " << nodeProduct << std::endl;

  fclose(f);
}  
  

namespace {

  int compare(const void* a, const void* b) {
    std::string str1, str2;
    
    str1 = move_to_string(((MoveStack *)a)->move);
    str2 = move_to_string(((MoveStack *)b)->move);
    return str1.compare(str2);
  }


  Color random_color() {
    return Color(genrand_int32() % 2);
  }
  

  Square random_square() {
    return Square(genrand_int32() % 64);
  }

  
  Square random_empty_square(const Position& pos) {
    Square s;
    for(s = random_square(); !pos.square_is_empty(s); s = random_square());
    return s;
  }


  PieceType random_promoted_piece_type() {
    return PieceType((genrand_int32() % 4) + 2);
  }
  
}
