// Copyright (C) 2004-2007 Tord Romstad

// This file is part of Glaurung.
//
// Glaurung is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//  
// Glaurung is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Glaurung; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA


////
//// Includes
////

#include <cstdarg>
#include <sstream>
#include <vector>

#include "ucioption.h"


////
//// Local definitions
////

namespace {

  ///
  /// Types
  ///

  enum UCIOptionType {
    SPIN, COMBO, CHECK, STRING, BUTTON
  };

 
  class UCIOption {

  public:
    UCIOption(const std::string& aName, const std::string& aDefaultValue,
              UCIOptionType aType, int min, int max, ...);
    void print() const;
    const std::string get_name() const;
    const std::string get_value() const;
    void set_value(const std::string& newValue);

  private:
    std::string name;
    std::string defaultValue;
    std::string currentValue;
    UCIOptionType type;
    int minValue, maxValue;
    std::vector <std::string> comboValues;

  };


  ///
  /// Variables
  ///

  std::vector <UCIOption *> UCIOptions;

  ///
  /// Functions
  ///

  UCIOption *option_with_name(const std::string& optionName);

}


////
//// Functions
////

void init_uci_options() {
  UCIOptions.push_back(new UCIOption("Hash", "32", SPIN, 4, 1024));
  UCIOptions.push_back(new UCIOption("Ponder", "true", CHECK, 0, 0));
  UCIOptions.push_back(new UCIOption("OwnBook", "true", CHECK, 0, 0));
  UCIOptions.push_back(new UCIOption("UCI_ShowCurrLine", "false", CHECK,
				     0, 0));
  UCIOptions.push_back(new UCIOption("UCI_Chess960", "false", CHECK, 0, 0));
  UCIOptions.push_back(new UCIOption("UCI_EngineAbout", 
                                     "Glaurung by Tord Romstad, see http://www.glaurungchess.com",
                                     STRING, 0, 0));
  UCIOptions.push_back(new UCIOption("Threads", "1", SPIN, 1, 4));
}


void print_uci_options() {
  for(unsigned i = 0; i < UCIOptions.size(); i++)
    UCIOptions.at(i)->print();
}


bool get_option_value_bool(const std::string& optionName) {
  UCIOption *u = option_with_name(optionName);
  if(u != NULL && u->get_value() == "true")
    return true;
  else
    return false;
}


int get_option_value_int(const std::string& optionName) {
  UCIOption *u = option_with_name(optionName);
  std::string value = u->get_value();
  std::istringstream i(value);
  int intValue;

  if(i >> intValue)
    return intValue;

  return 0;
}


const std::string get_option_value_string(const std::string& optionName) {
  UCIOption *u = option_with_name(optionName);
  return u->get_value();
}


bool button_was_pressed(const std::string& buttonName) {
  // Not implemented yet. FIXME
  return false;
}


void set_option_value(const std::string& optionName,
                      const std::string& newValue) {
  UCIOption *u = option_with_name(optionName);

  if(u != NULL)
    u->set_value(newValue);
  else
    std::cout << "No such option: " << optionName << std::endl;
}


void push_button(const std::string& buttonName) {
  // Not implemented yet. FIXME
}


namespace {

  UCIOption::UCIOption(const std::string& aName, 
                       const std::string& aDefaultValue,
                       UCIOptionType aType, int min, int max, ...) {
    name.insert(0, aName);
    defaultValue.insert(0, aDefaultValue);
    currentValue.insert(0, aDefaultValue);
    minValue = min; 
    maxValue = max;
    type = aType;

    if(type == COMBO) {
      va_list arglist;
      char *str;
      va_start(arglist, max);
      do {
        str = va_arg(arglist, char *);
        if(str != NULL) 
          comboValues.push_back(std::string(str));
      } while(str != NULL);
      va_end(arglist);
    }
  }


  void UCIOption::print() const {
    const std::string optionTypeName[] = {
      "spin", "combo", "check", "string", "button"
    };
    std::cout << "option name " << name << " type " << optionTypeName[type];

    if(type != BUTTON) {
      std::cout << " default " << defaultValue;

      if(type == SPIN)
        std::cout << " min " << minValue << " max " << maxValue;
      else if(type == COMBO)
        for(unsigned i = 0; i < comboValues.size(); i++)
          std::cout << " var " << comboValues.at(i);
    }

    std::cout << std::endl;
  }


  const std::string UCIOption::get_name() const {
    return name;
  }


  const std::string UCIOption::get_value() const {
    return currentValue;
  }


  void UCIOption::set_value(const std::string& newValue) {
    currentValue.clear();
    currentValue.insert(0, newValue);
  }


  UCIOption *option_with_name(const std::string& optionName) {
    for(unsigned i = 0; i < UCIOptions.size(); i++)
      if(UCIOptions.at(i)->get_name() == optionName)
        return UCIOptions.at(i);
    return NULL;
  }

}
